import boto3
from botocore.exceptions import NoCredentialsError
import sys
from google_images_download import google_images_download
import os
import shutil
from metal.utils.production_util import create_training_data as create_training_data_
from dotenv import load_dotenv
import os
from os.path import join, dirname

response = google_images_download.googleimagesdownload()

class Data_loader(object):
    """docstring for data_loader."""

    def __init__(self, path=None):
        super(Data_loader, self).__init__()
        self.output_directory = path

    def init_s3(self, env_path=None):

        load_dotenv(env_path)
        AWSAccessKeyId = os.getenv("AWSAccessKeyId")
        AWSSecretKey = os.getenv("AWSSecretKey")
        region = os.getenv("region")

        self.s3 = boto3.client('s3', aws_access_key_id=AWSAccessKeyId,
                          aws_secret_access_key=AWSSecretKey)

    def upload_to_awsS3(self, local_file, bucket, s3_file):
        try:
            self.s3.upload_file(local_file, bucket, s3_file)
            print("Upload Successful")
            return True
        except FileNotFoundError:
            print("The file was not found")
            return False
        except NoCredentialsError:
            print("Credentials not available")
            return False

    def download(self, query, image_directory, format="jpg",print_urls=True, limit=100, size="medium", aspect_ratio="panoramic",type_="photo", output_directory=None):
        #bug: should check if downloads dir alrady exist
        #need to be instacnce var
        if output_directory == None:
            self.output_directory

        arguments = {"keywords": query,
             "format": format,
             "limit":limit,
             "print_urls":print_urls,
             "size": size,
             "aspect_ratio": aspect_ratio,
             "type":type_,
             "output_directory":self.output_directory,
             "image_directory":image_directory}
        try:
            response.download(arguments)
        # Handling File NotFound Error
        except FileNotFoundError:
            arguments = {"keywords": query,
                 "format": format,
                 "limit":limit,
                 "print_urls":print_urls,
                 "size": size,
                 "aspect_ratio": aspect_ratio,
                 "type":type_,
                 "output_directory":self.output_directory,
                 "image_directory":image_directory}
            # Providing arguments for the searched query
            try:
                # Downloading the photos based
                # on the given arguments
                response.download(arguments)
            except:
                pass

    def create_training_data(self,classes,img_size=50):
        return create_training_data_(img_size=img_size, classes=classes, data_dir=self.output_directory, color=None)

    def delete_dir(self):
         shutil.rmtree(self.output_directory)



import numpy as np
from metal.layers.layer import PoolingLayer

class MaxPooling2D(PoolingLayer):
    def _pool_forward(self, X_col):
        arg_max = np.argmax(X_col, axis=0).flatten()
        output = X_col[arg_max, range(arg_max.size)]
        self.cache = arg_max
        return output

    def _pool_backward(self, accum_grad):
        accum_grad_col = np.zeros((np.prod(self.pool_shape), accum_grad.size))
        arg_max = self.cache
        accum_grad_col[arg_max, range(accum_grad.size)] = accum_grad
        return accum_grad_col




import cupy as cp
from autogradgpu.node import Node


class Parameter(Node):
    def __init__(self, data=None, requires_grad=True, depends_on=None, create_array=None, name=None):
        self.name = name
        if create_array is not None:
            cp.random.seed(create_array[0])
            data_ = cp.random.randn(*create_array)
        elif data is not None:
            data_ = data
        super().__init__(data=data_, requires_grad=requires_grad, depends_on=depends_on)

    def __repr__(self) -> str:
        return f"Parameter({self.data}, requires_grad={self.requires_grad})"




from typing import List
import cupy as cp
from autogradgpu.node import Node
from autogradgpu.dependency import Dependency


class Tensor(Node):
    """docstring for Tensor."""
    def __init__(self, data: cp.ndarray, requires_grad: bool = False, depends_on: List[Dependency] = None, name=None):
        super().__init__(data=data,requires_grad=requires_grad, depends_on=depends_on)
        self.name = name

    def __repr__(self) -> str:
        return f"Tensor({self.data}, requires_grad={self.requires_grad})"



import numpy as np
from autograd.tensor import Tensor
from autograd.parameter import Parameter
from metal.module import Module
from autograd.dependency import Dependency
import math
import copy
from metal.layers.layer import Layer


class BatchNormalization(Layer):
    __slots__ = ('momentum','trainable','eps','running_mean','running_var','gamma','beta','gamma_opt','beta_opt','type','running_mean','running_var','X_centered','stddev_inv')
    """
    Batch normalization.
    """
    def __init__(self, momentum=0.99):
        self.momentum = momentum
        self.trainable = True
        self.eps = 0.01
        self.running_mean = None
        self.running_var = None
        self.load_params_ = False

    def initialize(self, optimizer):
        # Initialize the parameters
        self.gamma = Parameter(np.ones(self.input_shape))
        self.beta = Parameter(np.zeros(self.input_shape))
        # parameter optimizers
        self.gamma_opt  = copy.copy(optimizer)
        self.beta_opt = copy.copy(optimizer)


    def forward_pass(self, x, training=True):
        # Initialize running mean and variance if first run
        X = x.data
        self.type = type(x)
        requires_grad = x.requires_grad or self.gamma.requires_grad or self.beta.requires_grad
        depends_on: List[Dependency] = []

        if self.running_mean is None:
            self.running_mean = np.mean(X, axis=0)
            self.running_var = np.var(X, axis=0)

        if training and self.trainable:
            mean = np.mean(X, axis=0)
            var = np.var(X, axis=0)
            self.running_mean = self.momentum * self.running_mean + (1 - self.momentum) * mean
            self.running_var = self.momentum * self.running_var + (1 - self.momentum) * var
        else:
            mean = self.running_mean
            var = self.running_var

        # Statistics saved for backward pass
        self.X_centered = X - mean
        self.stddev_inv = 1 / np.sqrt(var + self.eps)

        X_norm = self.X_centered * self.stddev_inv
        data = self.gamma.data * X_norm + self.beta.data
        if training:
            if requires_grad:
                if self.gamma.requires_grad:
                    depends_on.append(Dependency(self.gamma, self.grad_gamma_batchNorm))
                if self.beta.requires_grad:
                    depends_on.append(Dependency(self.beta, self.grad_beta_batchNorm))
                if x.requires_grad:
                    depends_on.append(Dependency(x, self.grad_x_batchNorm))
            else:
                depends_on = []
        else:
            depends_on = []
        return self.type(data=data,requires_grad=requires_grad,depends_on=depends_on)


    def grad_gamma_batchNorm(self, grad):
        # Save parameters used during the forward pass
        X_norm = self.X_centered * self.stddev_inv
        grad_gamma = np.sum(grad * X_norm, axis=0)
        return grad_gamma


    def grad_beta_batchNorm(self, grad):
        # Save parameters used during the forward pass
        grad_beta = np.sum(grad, axis=0)
        return grad_beta

    def grad_x_batchNorm(self, grad):
        # Save parameters used during the forward pass
        gamma = self.gamma.data
        batch_size = grad.shape[0]
        # The gradient of the loss with respect to the layer inputs (use weights and statistics from forward pass)
        accum_grad = (1 / batch_size) * gamma * self.stddev_inv * (
            batch_size * grad
            - np.sum(grad, axis=0)
            - self.X_centered * self.stddev_inv**2 * np.sum(grad * self.X_centered, axis=0)
            )

        return accum_grad

    def update_pass(self):
        # If the layer is trainable the parameters are updated
        if self.trainable:
            self.gamma = self.gamma_opt.update(self.gamma)
            self.beta = self.beta_opt.update(self.beta)
        # clear the gradients
        self.zero_grad()

    def output_shape(self):
        return self.input_shape




import numpy as np
from autograd.dependency import Dependency
import math

class Slice(object):
    __slots__ = ('type','t','name','idxs')

    """docstring for Slice."""
    def __init__(self, t, idxs, name_child=None):
        self.type = type(t)
        self.t = t
        self.name = name_child
        self.idxs = idxs

    def _slice(self):
        data = self.t.data[self.idxs]
        requires_grad = self.t.requires_grad
        if requires_grad:
            depends_on = [Dependency(self.t, self.grad_slice)]
        else:
            depends_on = []
        return self.type(data, requires_grad, depends_on, name=self.name)

    def grad_slice(self, grad: np.ndarray) -> np.ndarray:
        old_shape = self.t.shape
        new_grad = np.zeros(old_shape, dtype=np.float32)
        new_grad[self.idxs] = grad
        return new_grad


class T(object):
    __slots__ = ('type','t','axis','name')

    """docstring for Transpose."""
    def __init__(self, t, axis, name_child=None):
        self.type = type(t)
        self.t = t
        self.axis = axis
        self.name = name_child

    def _T(self):
        data = self.t.data.transpose(*self.axis)
        requires_grad = self.t.requires_grad
        if requires_grad:
            depends_on = [Dependency(self.t, self.grad_t)]
        else:
            depends_on = []
        return self.type(data, requires_grad, depends_on, name=self.name)

    def grad_t(self, grad: np.ndarray):
        return grad.reshape(self.t.data.shape)

class Reshape(object):
    __slots__ = ('type','new_shape','t','name')

    """docstring for Reshape."""
    def __init__(self, t, newshape, name_child=None):
        super(Reshape, self).__init__()
        self.type = type(t)
        self.new_shape = newshape
        self.t = t
        self.name = name_child


    def _reshape(self):
        data = self.t.data.reshape(*self.new_shape)
        requires_grad = self.t.requires_grad
        if requires_grad:
            depends_on = [Dependency(self.t, self.grad_reshape)]
        else:
            depends_on = []
        return self.type(data, requires_grad, depends_on, name=self.name)

    def grad_reshape(self, grad: np.array):
        old_shape = self.t.shape
        return grad.reshape(*old_shape)
        #return grad.transpose([x for x in range(self.t.data.ndim)])

class Pad(object):
    __slots__ = ('type','t','pad','name')

    """docstring for Pad."""
    def __init__(self, t, pad, name_child=None):
        super(Pad, self).__init__()
        self.type = type(t)
        self.t = t
        self.pad = pad
        self.name = name_child

    def _pad(self):
        data = np.pad(self.t.data, self.pad[0], mode=self.pad[1])
        requires_grad = self.t.requires_grad
        if requires_grad:
            depends_on = [Dependency(self.t, self.grad_pad)]
        else:
            depends_on = []
        return self.type(data, requires_grad, depends_on, name=self.name )

    def grad_pad(self, grad: np.array):
        slices = []
        for c in self.pad[0]:
            e = None if c[1] == 0 else -c[1]
            slices.append(slice(c[0], e))
        return grad[tuple(slices)]



import numpy as np
from autograd.tensor import Tensor
from autograd.dependency import Dependency

# Collection of activation functions
# Reference: https://en.wikipedia.org/wiki/Activation_function

class Sigmoid(object):
    __slots__ = ('TYPE','data')
    """docstring for Sigmoid."""

    def __call__(self, x, training):
        self.TYPE = type(x)
        requires_grad = x.requires_grad
        self.data =  1 / (1 + np.exp(-x.data))
        if training:
            if requires_grad:
                depends_on = [Dependency(x, self.grad_sigmoid)]
            else:
                depends_on = []
        else:
            depends_on = []
        return self.TYPE(data=self.data,requires_grad=requires_grad,depends_on=depends_on)

    def grad_sigmoid(self, grad):
        return (self.data * (1 - self.data)) * grad


class TanH(object):
    __slots__ = ('TYPE','data')
    """docstring for TanH."""

    def __call__(self, x):
        self.TYPE = type(x)
        requires_grad = x.requires_grad
        self.data = np.tanh(x.data)
        if requires_grad:
            depends_on = [Dependency(x, self.grad_tanh)]
        else:
            depends_on = []
        return self.TYPE(data=self.data,requires_grad=requires_grad,depends_on=depends_on)

    def grad_tanh(self, grad):
        return (1 - self.data**2) * grad



class ReLU(object):
    __slots__ = ('type','x')

    """docstring for ReLU."""

    def __call__(self,x,training):
        self.x = x.data
        self.type = type(x)
        requires_grad = x.requires_grad

        data  = self.x * (self.x > 0)
        if training:
            if requires_grad:
                depends_on = [Dependency(x, self.grad_relu)]
            else:
                depends_on = []
        else:
            depends_on = []
        return self.type(data=data,requires_grad=requires_grad,depends_on=depends_on)

    def grad_relu(self, grad):
        return grad * (self.x>0)


class Softmax(object):
    __slots__ = ('TYPE','x')

    """docstring for Softmax."""

    def __call__(self, x, training):
        self.x = x.data
        self.TYPE = type(x)
        requires_grad = x.requires_grad

        e_x = np.exp(self.x - np.max(self.x, axis=-1, keepdims=True))
        e_x = e_x / np.sum(e_x, axis=-1, keepdims=True)
        if training:
            if requires_grad:
                depends_on = [Dependency(x, self.grad_softmax)]
            else:
                depends_on = []
        else:
            depends_on = []
        return self.TYPE(data=e_x,requires_grad=requires_grad,depends_on=depends_on)

    def grad_softmax(self, grad):
        e_x = np.exp(self.x - np.max(self.x, axis=-1, keepdims=True))
        e_x = e_x / np.sum(e_x, axis=-1, keepdims=True)
        return grad * (e_x * (1 - e_x))



"""
class LeakyReLU():
    def __call__(self, x, alpha=0.2):
        y1 = (x * (x.data > 0))
        y2 = (x * alpha * (x.data <= 0))
        return y1 + y2
"""

"""
class ReLU_():
    def __call__(self,x_):
        x = x_.data
        self.type = type(x_)
        requires_grad = x_.requires_grad
        if requires_grad:
            depends_on = [Dependency(x_, self.grad_relu)]
        else:
            depends_on = []
        return self.type(data=np.where(x >= 0, x, 0),requires_grad=requires_grad,depends_on=depends_on)

    def grad_relu(self, x):
        print(np.where(x >= 0, 1, 0))
        return np.where(x >= 0, 1, 0)
"""

# https://eli.thegreenplace.net/2016/the-softmax-function-and-its-derivative/
"""
class Softmax_():
    #docstring for Softmax.
    def __call__(self, x):
        shiftx = x - x.max()
        exps = shiftx.exp()
        return exps / exps.sum()
"""



import math
import numpy as np


# Method which calculates the padding based on the specified output shape and the
# shape of the filters
def determine_padding(filter_shape, output_shape="same"):

    # No padding
    if output_shape == "valid":
        return (0, 0), (0, 0)
    # Pad so that the output shape is the same as input shape (given that stride=1)
    elif output_shape == "same":
        filter_height, filter_width = filter_shape

        # Derived from:
        # output_height = (height + pad_h - filter_height) / stride + 1
        # In this case output_height = height and stride = 1. This gives the
        # expression for the padding below.
        pad_h1 = int(math.floor((filter_height - 1)/2))
        pad_h2 = int(math.ceil((filter_height - 1)/2))
        pad_w1 = int(math.floor((filter_width - 1)/2))
        pad_w2 = int(math.ceil((filter_width - 1)/2))

        return (pad_h1, pad_h2), (pad_w1, pad_w2)


# Reference: CS231n Stanford
def get_im2col_indices(images_shape, filter_shape, padding, stride=1):
    # First figure out what the size of the output should be
    batch_size, channels, height, width = images_shape
    filter_height, filter_width = filter_shape
    pad_h, pad_w = padding
    out_height = int((height + np.sum(pad_h) - filter_height) / stride + 1)
    out_width = int((width + np.sum(pad_w) - filter_width) / stride + 1)

    i0 = np.repeat(np.arange(filter_height), filter_width)
    i0 = np.tile(i0, channels)
    i1 = stride * np.repeat(np.arange(out_height), out_width)
    j0 = np.tile(np.arange(filter_width), filter_height * channels)
    j1 = stride * np.tile(np.arange(out_width), out_height)
    i = i0.reshape(-1, 1) + i1.reshape(1, -1)
    j = j0.reshape(-1, 1) + j1.reshape(1, -1)

    k = np.repeat(np.arange(channels), filter_height * filter_width).reshape(-1, 1)

    return (k, i, j)


# Method which turns the image shaped input to column shape.
# Used during the forward pass.
# Reference: CS231n Stanford
def image_to_column(images, filter_shape, stride, output_shape='same'):
    filter_height, filter_width = filter_shape

    pad_h, pad_w = determine_padding(filter_shape, output_shape)

    # Add padding to the image
    images_padded = np.pad(images, ((0, 0), (0, 0), pad_h, pad_w), mode='constant')

    # Calculate the indices where the dot products are to be applied between weights
    # and the image
    k, i, j = get_im2col_indices(images.shape, filter_shape, (pad_h, pad_w), stride)

    # Get content from image at those indices
    cols = images_padded[:, k, i, j]
    channels = images.shape[1]
    # Reshape content into column shape
    cols = cols.transpose(1, 2, 0).reshape(filter_height * filter_width * channels, -1)
    return cols



# Method which turns the column shaped input to image shape.
# Used during the backward pass.
# Reference: CS231n Stanford
def column_to_image(cols, images_shape, filter_shape, stride, output_shape='same'):
    batch_size, channels, height, width = images_shape
    pad_h, pad_w = determine_padding(filter_shape, output_shape)
    height_padded = height + np.sum(pad_h)
    width_padded = width + np.sum(pad_w)
    images_padded = np.zeros((batch_size, channels, height_padded, width_padded))

    # Calculate the indices where the dot products are applied between weights
    # and the image
    k, i, j = get_im2col_indices(images_shape, filter_shape, (pad_h, pad_w), stride)

    cols = cols.reshape(channels * np.prod(filter_shape), -1, batch_size)
    cols = cols.transpose(2, 0, 1)

    # Add column content to the images at the indices
    np.add.at(images_padded, (slice(None), k, i, j), cols)
    # Return image without padding
    return images_padded[:, :, pad_h[0]:height+pad_h[0], pad_w[0]:width+pad_w[0]]



from __future__ import division
from itertools import combinations_with_replacement
import numpy as np
import math
import sys

def shuffle_data(X, y, seed=None):
    """ Random shuffle of the samples in X and y """
    if seed:
        np.random.seed(seed)
    idx = np.arange(X.shape[0])
    np.random.shuffle(idx)
    return X[idx], y[idx]


def batch_iterator(X, y=None, batch_size=64):
    """ Simple batch generator """
    n_samples = X.shape[0]
    for i in np.arange(0, n_samples, batch_size):
        begin, end = i, min(i+batch_size, n_samples)
        if y is not None:
            yield X[begin:end], y[begin:end]
        else:
            yield X[begin:end]


def divide_on_feature(X, feature_i, threshold):
    """ Divide dataset based on if sample value on feature index is larger than
        the given threshold """
    split_func = None
    if isinstance(threshold, int) or isinstance(threshold, float):
        split_func = lambda sample: sample[feature_i] >= threshold
    else:
        split_func = lambda sample: sample[feature_i] == threshold

    X_1 = np.array([sample for sample in X if split_func(sample)])
    X_2 = np.array([sample for sample in X if not split_func(sample)])

    return np.array([X_1, X_2])


def polynomial_features(X, degree):
    n_samples, n_features = np.shape(X)

    def index_combinations():
        combs = [combinations_with_replacement(range(n_features), i) for i in range(0, degree + 1)]
        flat_combs = [item for sublist in combs for item in sublist]
        return flat_combs

    combinations = index_combinations()
    n_output_features = len(combinations)
    X_new = np.empty((n_samples, n_output_features))

    for i, index_combs in enumerate(combinations):
        X_new[:, i] = np.prod(X[:, index_combs], axis=1)

    return X_new


def get_random_subsets(X, y, n_subsets, replacements=True):
    """ Return random subsets (with replacements) of the data """
    n_samples = np.shape(X)[0]
    # Concatenate x and y and do a random shuffle
    X_y = np.concatenate((X, y.reshape((1, len(y))).T), axis=1)
    np.random.shuffle(X_y)
    subsets = []

    # Uses 50% of training samples without replacements
    subsample_size = int(n_samples // 2)
    if replacements:
        subsample_size = n_samples      # 100% with replacements

    for _ in range(n_subsets):
        idx = np.random.choice(
            range(n_samples),
            size=np.shape(range(subsample_size)),
            replace=replacements)
        X = X_y[idx][:, :-1]
        y = X_y[idx][:, -1]
        subsets.append([X, y])
    return subsets


def normalize(X, axis=-1, order=2):
    """ Normalize the dataset X """
    l2 = np.atleast_1d(np.linalg.norm(X, order, axis))
    l2[l2 == 0] = 1
    return X / np.expand_dims(l2, axis)


def standardize(X):
    """ Standardize the dataset X """
    X_std = X
    mean = X.mean(axis=0)
    std = X.std(axis=0)
    for col in range(np.shape(X)[1]):
        if std[col]:
            X_std[:, col] = (X_std[:, col] - mean[col]) / std[col]
    # X_std = (X - X.mean(axis=0)) / X.std(axis=0)
    return X_std


def train_test_split(X, y, test_size=0.5, shuffle=True, seed=None):
    """ Split the data into train and test sets """
    if shuffle:
        X, y = shuffle_data(X, y, seed)
    # Split the training data from test data in the ratio specified in
    # test_size
    split_i = len(y) - int(len(y) // (1 / test_size))
    X_train, X_test = X[:split_i], X[split_i:]
    y_train, y_test = y[:split_i], y[split_i:]

    return X_train, X_test, y_train, y_test


def k_fold_cross_validation_sets(X, y, k, shuffle=True):
    """ Split the data into k sets of training / test data """
    if shuffle:
        X, y = shuffle_data(X, y)

    n_samples = len(y)
    left_overs = {}
    n_left_overs = (n_samples % k)
    if n_left_overs != 0:
        left_overs["X"] = X[-n_left_overs:]
        left_overs["y"] = y[-n_left_overs:]
        X = X[:-n_left_overs]
        y = y[:-n_left_overs]

    X_split = np.split(X, k)
    y_split = np.split(y, k)
    sets = []
    for i in range(k):
        X_test, y_test = X_split[i], y_split[i]
        X_train = np.concatenate(X_split[:i] + X_split[i + 1:], axis=0)
        y_train = np.concatenate(y_split[:i] + y_split[i + 1:], axis=0)
        sets.append([X_train, X_test, y_train, y_test])

    # Add left over samples to last set as training samples
    if n_left_overs != 0:
        np.append(sets[-1][0], left_overs["X"], axis=0)
        np.append(sets[-1][2], left_overs["y"], axis=0)

    return np.array(sets)


def to_categorical(x, n_col=None):
    """ One-hot encoding of nominal values """
    if not n_col:
        n_col = np.amax(x) + 1
    one_hot = np.zeros((x.shape[0], n_col))
    one_hot[np.arange(x.shape[0]), x] = 1
    return one_hot


def to_nominal(x):
    """ Conversion from one-hot encoding to nominal """
    return np.argmax(x, axis=1)


def make_diagonal(x):
    """ Converts a vector into an diagonal matrix """
    m = np.zeros((len(x), len(x)))
    for i in range(len(m[0])):
        m[i, i] = x[i]
    return m



from metal.nn import NeuralNetwork
import matplotlib.pyplot as plt

class ConvNet(NeuralNetwork):
    """docstring for CovNet."""
    
    def __init__(self, optimizer, loss, validation_data=None):
        super(ConvNet, self).__init__(optimizer, loss, validation_data=validation_data)

    def eval(self, X_test, y_test):
        train_err, val_err = self.errors['training'], self.errors["validation"]
        # Training and validation error plot
        n = len(train_err)
        training, = plt.plot(range(n), train_err, label="Training Error")
        validation, = plt.plot(range(n), val_err, label="Validation Error")
        plt.legend(handles=[training, validation])
        plt.title("Error Plot")
        plt.ylabel('Error')
        plt.xlabel('Iterations')
        plt.show()
        _, accuracy = self.test_on_batch(X_test, y_test)
        print ("Accuracy:", accuracy)

    def predict(self, X):
        pred = self._forward_pass(X, training=False)
        list_pred = pred.data.flatten().tolist()
        return list_pred.index(max(list_pred))



from __future__ import division
from typing import List, NamedTuple, Callable, Optional, Union
import numpy as np
from autograd.math_operations import Sum, Add, MatMul, Mul, Sub, Div, Neg, Exp, Max
from autograd.tensor_modifications import Slice, T, Reshape, Pad
from autograd.dependency import Dependency
from autograd.engin import  autograd


Arrayable = Union[float, list, np.ndarray]


def ensure_array(arrayable: Arrayable) -> np.ndarray:
    if isinstance(arrayable, np.ndarray):
        arrayable =  arrayable.astype(np.float32, copy=False)
        return arrayable
    else:
        return np.array(arrayable).astype(np.float32, copy=False)


Nodeable = Union["Node", float, np.ndarray]


def ensure_Type(Nodeable: Nodeable, t) -> "Node":
    if isinstance(Nodeable, Node):
        return Nodeable
    else:
        TYPE = type(t)
        return TYPE(Nodeable, requires_grad=False)



class Node(object):
    __slots__ = ('_data','requires_grad','depends_on','shape','grad','name','id')
    def __init__(self,data: Arrayable,requires_grad: bool = False,depends_on: List[Dependency] = None, id = None, name=None) -> None:
        self._data = ensure_array(data)
        self.requires_grad = requires_grad
        self.depends_on = depends_on or []
        self.shape = self._data.shape
        self.grad: Optional["Node"] = None
        self.name = name
        if id is None:
            id = np.random.randint(0, 100_000)
        self.id = id
        if self.requires_grad:
            self.zero_grad()

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, new_data: np.ndarray) -> None:
        self._data = new_data
        # Setting the data manually means we invalidate the gradient.
        self.grad = None

    def zero_grad(self) -> None:
        self.type = type(self)
        self.grad = self.type(np.zeros_like(self.data, dtype=np.float32),requires_grad=False)

    def __repr__(self) -> str:
        return f"Node({self.data}, requires_grad={self.requires_grad})"

    def __add__(self, other) -> "Node":
        """gets called if I do t + other"""
        return Add(self,ensure_Type(other,self))._add()

    def __radd__(self, other) -> "Node":
        """gets called if I do other + t"""
        return Add(ensure_Type(other,self), self)._add()

    def __iadd__(self, other) -> "Node":
        """when we do t += other"""
        self.data = self.data + ensure_Type(other,self).data
        return self

    def __isub__(self, other) -> "Node":
        """when we do t -= other"""
        self.data = self.data - ensure_Type(other,self).data
        return self

    def __imul__(self, other) -> "Node":
        """when we do t *= other"""
        self.data = self.data * ensure_Type(other,self).data
        return self

    def __mul__(self, other) -> "Node":
        return Mul(self, ensure_Type(other,self))._mul()

    def __rmul__(self, other) -> "Node":
        return Mul(ensure_Type(other,self), self)._mul()

    def __matmul__(self, other) -> "Node":
        return MatMul(self, other)._matmul()

    def __neg__(self) -> "Node":
        return Neg(self)._neg()

    def __sub__(self, other) -> "Node":
        return Sub(self, ensure_Type(other,self))._sub()

    def __rsub__(self, other) -> "Node":
        return Sub(self, ensure_Type(other,self))._sub()

    def __getitem__(self, idxs) -> "Node":
        return Slice(self, idxs)._slice()

    def __truediv__(self, other) -> "Node":
        """gets called if I do t / other"""
        return Div(self, ensure_Type(other,self))._div()

    def __rtruediv__(self, other) -> "Node":
        """gets called if I do other / t"""
        return Div(ensure_Type(other,self), self)._div()

    def sum(self,axis=None) -> "Node":
        return Sum(self,axis)._sum()

    def T(self,*axis) -> "Node":
        return T(self,axis)._T()

    def pad(self,*pad) -> "Node":
        return Pad(self,pad)._pad()

    def reshape(self,*newshape)-> "Node":
        return Reshape(self,newshape)._reshape()

    def exp(self)->"Node":
        return Exp(self)._exp()

    def max(self)->"Node":
        return Max(self)._max()

    def backward(self, grad: "Node" = None) -> None:
        autograd(self).backward(grad) # apply backward function wrapping the output gardent



import numpy as np
from autograd.dependency import Dependency
import math
from metal.utils.layer_data_manipulations import determine_padding, get_im2col_indices

"""
Utility file for defining functions or a classes that will forward prop and backward prop within metal.
"""

class IMG2COL(object):
    """docstring for IMG2COL."""

    def __init__(self, images, filter_shape, stride, output_shape='same'):
        super(IMG2COL, self).__init__()
        self.type = type(images)
        self.images = images
        self.filter_shape = filter_shape
        self.stride = stride
        self.output_shape = output_shape
        self.images_shape = images.shape


    def image_to_column(self):
        depends_on: List[Dependency] = []
        filter_height, filter_width = self.filter_shape

        pad_h, pad_w = determine_padding(self.filter_shape, self.output_shape)

        # Add padding to the image
        images_padded = np.pad(self.images.data, ((0, 0), (0, 0), pad_h, pad_w), mode='constant')

        # Calculate the indices where the dot products are to be applied between weights
        # and the image
        k, i, j = get_im2col_indices(self.images_shape, self.filter_shape, (pad_h, pad_w), self.stride)

        # Get content from image at those indices
        cols = images_padded[:, k, i, j]
        channels = self.images_shape[1]
        # Reshape content into column shape
        cols = cols.transpose(1, 2, 0).reshape(filter_height * filter_width * channels, -1)

        requires_grad=self.images.requires_grad
        if self.images.requires_grad:
            depends_on.append(Dependency(self.images, self.column_to_image))
        return self.type(data=cols,requires_grad=requires_grad,depends_on=depends_on)

    def column_to_image(self, grad):
        batch_size, channels, height, width = self.images_shape
        pad_h, pad_w = determine_padding(self.filter_shape, self.output_shape)
        height_padded = height + np.sum(pad_h)
        width_padded = width + np.sum(pad_w)
        images_padded = np.zeros((batch_size, channels, height_padded, width_padded))

        # Calculate the indices where the dot products are applied between weights
        # and the image
        k, i, j = get_im2col_indices(self.images_shape, self.filter_shape, (pad_h, pad_w), self.stride)

        cols = grad.reshape(channels * np.prod(self.filter_shape), -1, batch_size)
        cols = cols.transpose(2, 0, 1)
        # Add column content to the images at the indices
        np.add.at(images_padded, (slice(None), k, i, j), cols)

        # Return image without padding
        return images_padded[:, :, pad_h[0]:height+pad_h[0], pad_w[0]:width+pad_w[0]]

class Trans(object):
    """docstring for Trans."""

    def __init__(self, t, axis_f, axis_b):
        super(Trans, self).__init__()
        self.type = type(t)
        self.t = t
        self.axis_f = axis_f
        self.axis_b = axis_b

    def trans(self):
        data = self.t.data.transpose(*self.axis_f)
        requires_grad = self.t.requires_grad
        depends_on: List[Dependency] = []
        if requires_grad:
            depends_on.append(Dependency(self.t, self.Trans_grad))
        return self.type(data=data,requires_grad=requires_grad,depends_on=depends_on)

    def Trans_grad(self,grad):
        return grad.transpose(*self.axis_b)



from __future__ import print_function, division
from terminaltables import AsciiTable
import cupy as cp
import progressbar
from metalgpu.utils import batch_iterator
from metalgpu.utils.misc import bar_widgets
import matplotlib.pyplot as plt

class NeuralNetwork(object):
    __slots__ = ('optimizer','layers','errors','loss_function','progressbar','val_set','trainable')
    """Neural Network. Deep Learning base model.
    Parameters:
    -----------
    optimizer: class
        The weight optimizer that will be used to tune the weights in order of minimizing
        the loss.
    loss: class
        Loss function used to measure the model's performance. SquareLoss or CrossEntropy.
    validation: tuple
        A tuple containing validation data and labels (X, y)
    """
    def __init__(self, optimizer, loss, validation_data=None):
        self.optimizer = optimizer
        self.layers = []
        self.errors = {"training": [], "validation": []}
        self.acc = {"training": [], "validation": []}
        self.loss_function = loss
        self.progressbar = progressbar.ProgressBar(widgets=bar_widgets)
        self.val_set = None
        if validation_data:
            X, y = validation_data
            self.val_set = {"X": X, "y": y}

    def set_trainable(self, trainable):
        """ Method which enables freezing of the weights of the network's layers. """
        for layer in self.layers:
            layer.trainable = trainable

    def add(self, layer):
        """ Method which adds a layer to the neural network """
        # If this is not the first layer added then set the icput shape
        # to the output shape of the last added layer
        if self.layers:
            layer.set_icput_shape(shape=self.layers[-1].output_shape())

        # If the layer has weights that needs to be initialized
        if hasattr(layer, 'load_params'):
            if layer.load_params_ == True:
                layer.load_optimzer(optimizer=self.optimizer)
            else:
                layer

        if hasattr(layer, 'initialize'):
            if layer.load_params_ == False:
                layer.initialize(optimizer=self.optimizer)
            else:
                layer

        # Add layer to the network
        self.layers.append(layer)

    def test_on_batch(self, X, y):
        """ Evaluates the model over a single batch of samples """
        y_pred = self._forward_pass(X, training=False)
        loss = self.loss_function(y, y_pred).loss(training=False)
        acc = self.loss_function(y, y_pred).acc()

        return loss, acc

    def train_on_batch(self, X, y):
        """ Single gradient update over one batch of samples """
        y_pred = self._forward_pass(X, training=True)
        loss = self.loss_function(y, y_pred).loss(training=True)
        acc = self.loss_function(y, y_pred).acc()
        #Calculate the gradient of the loss function wrt y_pred
        loss.sum().backward()
        #Update weights
        self._update_pass()

        return loss.data, acc

    def fit(self, X, y, n_epochs, batch_size):
        """ Trains the model for a fixed number of epochs """
        for _ in self.progressbar(range(n_epochs)):
            batch_acc = []
            batch_error = []
            for X_batch, y_batch in batch_iterator(X, y, batch_size=batch_size):
                loss, acc = self.train_on_batch(X_batch, y_batch)
                batch_error.append(loss)
                batch_acc.append(acc)
                print('train -->',acc)


            self.errors["training"].append(cp.mean(batch_error))
            self.acc["training"].append(cp.mean(batch_acc))

            if self.val_set is not None:
                val_loss, val_acc = self.test_on_batch(self.val_set["X"], self.val_set["y"])
                self.errors["validation"].append(val_loss.data)
                self.acc["validation"].append(val_acc.data)
                print('test ------>',val_acc)


        return self.errors["training"], self.errors["validation"]

    def _forward_pass(self, X, training):
        """ Calculate the output of the NN """
        layer_output = X
        for layer in self.layers:
            layer_output = layer.forward_pass(layer_output, training)
        return layer_output

    def _update_pass(self):
        """ Propagate the gradient 'backwards' and update the weights in each layer """
        for layer in reversed(self.layers):
            loss_grad = layer.update_pass()

    def summary(self, name="Model Summary"):
        # Print model name
        print (AsciiTable([[name]]).table)
        # Network icput shape (first layer's icput shape)
        print ("Icput Shape: %s" % str(self.layers[0].icput_shape))
        # Iterate through network and get each layer's configuration
        table_data = [["Layer Type", "Parameters", "Output Shape"]]
        tot_params = 0
        for layer in self.layers:
            layer_name = layer.layer_name()
            params = layer.parameters_()
            out_shape = layer.output_shape()
            table_data.append([layer_name, str(params), str(out_shape)])
            tot_params += params

        # Print network configuration table
        print (AsciiTable(table_data).table)
        print ("Total Parameters: %d\n" % tot_params)

    def predict(self, X):
        """ Use the trained model to predict labels of X """
        return self._forward_pass(X, training=False)



import os
import re
import sys
import platform
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion
from setuptools import find_packages

requires = [
    'six>=1.10.0'
]

setup(
    name='fullmetal',
    version='0.26',
    author='Miguel Myers',
    packages=find_packages(exclude=("tests",)),
    install_requires=requires,
    author_email='miguelmyers8@gmail.com',
    description='Deep Learning project',
    long_description='',
    include_package_data=True,
    zip_safe=False,
)




class autograd(object):
    __slots__ = ('type','t')
    """docstring for autograd."""

    def __init__(self, t):
        super(autograd, self).__init__()
        self.type = type(t)
        self.t = t

    def backward(self, grad: "Node" = None) -> None:
        assert self.t.requires_grad, "called backward on non-requires-grad Node"
        if grad is None:
            if self.t.shape == ():
                grad = self.type(1.0)
            else:
                raise RuntimeError("grad must be specified for non-0-Node")
        self.t.grad.data = self.t.grad.data + grad.data  # type: ignore
        for dependency in self.t.depends_on: #loop over the list
            backward_grad = dependency.grad_fn(grad.data) # apply gard fuction
            dependency.Node.backward(self.type(backward_grad)) # get current Node
                                                                      # apply backward function
                                                                      # wrapping the output gardent



import numpy as np
from metal.layers.layer import PoolingLayer

class AveragePooling2D(PoolingLayer):
    def _pool_forward(self, X_col):
        output = np.mean(X_col, axis=0)
        return output

    def _pool_backward(self, accum_grad):
        accum_grad_col = np.zeros((np.prod(self.pool_shape), accum_grad.size))
        accum_grad_col[:, range(accum_grad.size)] = 1. / accum_grad_col.shape[0] * accum_grad
        return accum_grad_col
